# ******************************************************************************
#  Copyright (c) 2021-2022. Kneron Inc. All rights reserved.                   *
# ******************************************************************************

import os
import sys
import argparse

PWD = os.path.dirname(os.path.abspath(__file__))
sys.path.insert(1, os.path.join(PWD, '..'))

from utils.ExampleHelper import get_device_usb_speed_by_port_id
import kp
import cv2

MODEL_FILE_PATH = os.path.join(PWD, '../../res/models/KL720/YoloV5s_640_640_3/models_720.nef')
IMAGE_FILE_PATH = os.path.join(PWD, '../../res/images/car_park_barrier_608x608.bmp')
LOOP_TIME = 50

def _connect_devices_with_product_id(product_id: kp.ProductId):
    print('[Check Device]')

    # scan devices
    _device_list = kp.core.scan_devices()

    # check Kneron device exist
    if _device_list.device_descriptor_number == 0:
        raise AttributeError('Error: no Kneron device !')

    _usb_port_ids = []

    # get device_descriptor of specified scan index
    for __idx, __device_descriptor in enumerate(_device_list.device_descriptor_list):
        if __device_descriptor.product_id == product_id and __device_descriptor.is_connectable:
            _usb_port_ids.append(__device_descriptor.usb_port_id)

    if len(_usb_port_ids) == 0:
        raise AttributeError('Error: no matched Kneron device of specified product ID !')

    print(' - success')

    try:
        print('[Connect Device]')
        _device_group = kp.core.connect_devices(usb_port_ids=_usb_port_ids)
        print(' - Success')
    except kp.ApiKPException as exception:
        print('Error: connect device fail, port ID = \'{}\', error msg: [{}]'.format(_usb_port_ids,
                                                                                     str(exception)))
        exit(0)

    return _device_group

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='KL720 Demo Generic Image Inference Example.')
    args = parser.parse_args()

    """
    check device USB speed (Recommend run KL720 at super speed)
    """
    try:
        if kp.UsbSpeed.KP_USB_SPEED_SUPER != get_device_usb_speed_by_port_id(usb_port_id=0):
            print('\033[91m' + '[Warning] Device is not run at super speed.' + '\033[0m')
    except Exception as exception:
        print('Error: check device USB speed fail, port ID = \'{}\', error msg: [{}]'.format(0,
                                                                                             str(exception)))
        exit(0)

    """
    connect the device
    """
    try:
        print('[Connect Device]')
        device_group = _connect_devices_with_product_id(product_id=kp.ProductId.KP_DEVICE_KL720)
        print(' - Success')
    except kp.ApiKPException as exception:
        print('Error: connect device fail, port ID = \'{}\', error msg: [{}]'.format(usb_port_id,
                                                                                     str(exception)))
        exit(0)

    """
    setting timeout of the usb communication with the device
    """
    print('[Set Device Timeout]')
    kp.core.set_timeout(device_group=device_group, milliseconds=5000)
    print(' - Success')

    """
    reboot the device
    """
    try:
        print('[Reboot Device]')
        kp.core.reset_device(device_group=device_group,
                             reset_mode=kp.ResetMode.KP_RESET_REBOOT,
                             sleep_secs=10)
        print(' - reboot command success')
    except kp.ApiKPException as exception:
        print('Error: reboot command failed, error msg: [{}]'.format(str(exception)))
        exit(0)

    """
    release the device group
    """
    try:
        print('[Disconnect Device]')
        kp.core.disconnect_devices(device_group=device_group)
        print(' - Success')
    except kp.ApiKPException as exception:
        print('Error: disconnect failed, error msg: [{}]'.format(str(exception)))
        exit(0)

    print(kp.core.scan_devices())


    """
    reconnect the device
    """
    try:
        print('[Reconnect Device]')
        device_group = _connect_devices_with_product_id(product_id=kp.ProductId.KP_DEVICE_KL720)
        print(' - Success')
    except kp.ApiKPException as exception:
        print('Error: reconnect device fail, port ID = \'{}\', error msg: [{}]'.format(usb_port_id,
                                                                                     str(exception)))
        exit(0)

    """
    setting timeout of the usb communication with the device
    """
    print('[Set Device Timeout]')
    kp.core.set_timeout(device_group=device_group, milliseconds=5000)
    print(' - Success')

    """
    upload model to device
    """
    try:
        print('[Upload Model]')
        model_nef_descriptor = kp.core.load_model_from_file(device_group=device_group,
                                                            file_path=MODEL_FILE_PATH)
        print(' - Success')
    except kp.ApiKPException as exception:
        print('Error: upload model failed, error = \'{}\''.format(str(exception)))
        exit(0)

    """
    prepare the image
    """
    print('[Read Image]')
    img = cv2.imread(filename=IMAGE_FILE_PATH)
    img_bgr565 = cv2.cvtColor(src=img, code=cv2.COLOR_BGR2BGR565)
    print(' - Success')

    """
    prepare generic image inference input descriptor
    """
    generic_inference_input_descriptor = kp.GenericImageInferenceDescriptor(
        model_id=model_nef_descriptor.models[0].id,
        inference_number=0,
        input_node_image_list=[
            kp.GenericInputNodeImage(
                image=img_bgr565,
                image_format=kp.ImageFormat.KP_IMAGE_FORMAT_RGB565,
                resize_mode=kp.ResizeMode.KP_RESIZE_ENABLE,
                padding_mode=kp.PaddingMode.KP_PADDING_CORNER,
                normalize_mode=kp.NormalizeMode.KP_NORMALIZE_KNERON
            )
        ]
    )

    """
    starting inference work
    """
    print('[Starting Inference Work]')
    print(' - Starting inference loop {} times'.format(LOOP_TIME))
    print(' - ', end='')
    for i in range(LOOP_TIME):
        try:
            kp.inference.generic_image_inference_send(device_group=device_group,
                                                      generic_inference_input_descriptor=generic_inference_input_descriptor)

            generic_raw_result = kp.inference.generic_image_inference_receive(device_group=device_group)
        except kp.ApiKPException as exception:
            print(' - Error: inference failed, error = {}'.format(exception))
            exit(0)

        print('.', end='', flush=True)
    print()

    """
    retrieve inference node output 
    """
    print('[Retrieve Inference Node Output ]')
    inf_node_output_list = []
    for node_idx in range(generic_raw_result.header.num_output_node):
        inference_float_node_output = kp.inference.generic_inference_retrieve_float_node(node_idx=node_idx,
                                                                                         generic_raw_result=generic_raw_result,
                                                                                         channels_ordering=kp.ChannelOrdering.KP_CHANNEL_ORDERING_CHW)
        inf_node_output_list.append(inference_float_node_output)

    print(' - Success')

    print('[Result]')
    print(inf_node_output_list)
